#===============================================================================
#
#   Code to reproduce G-effs of designs reported by anonymous authors of 
#    "Fast Computation of Highly G-optimal Exact Designs via Particle
#      Swarm Optimization".
#


#clear workspace
rm(list = ls())


#===============================================================================
# I) Several utility functions are needed to score a proposed design on the 
#  G-efficiency scale. These function follow in this section.

# 1. This function takes a design matrix(or vector) and expands 
#     it into a model matrix (or vector)
#   NOTE:
#    a) only supports second-order model for K = 1, 2, 3 factors

exp_Design_mat <- function(Xvec, N, n_factor){
  if(n_factor == 1){
    
    # design matrix
    dX <- data.frame(matrix(Xvec, 
                            byrow = F, 
                            nrow = N, 
                            ncol = n_factor))
    names(dX) <- "x1"
    # model matrix
    X <- model.matrix (~ 1 + x1 + I(x1^2), data = dX)
    
  }else if (n_factor == 2){
    
    # design matrix
    dX <- data.frame(matrix(Xvec, 
                            byrow = F, 
                            nrow = N, 
                            ncol = n_factor))
    names(dX) <- c("x1", "x2")
    # model matrix
    X <- model.matrix (~ 1 + x1 + x2 + x1*x2 +
                         I(x1^2) + I(x2^2), data = dX)
    
  }else if (n_factor == 3){
    
    # design matrix
    dX <- data.frame(matrix(Xvec, 
                            byrow = F, 
                            nrow = N, 
                            ncol = n_factor))
    names(dX) <- c("x1", "x2", "x3")
    # model matrix
    X <- model.matrix (~ 1 + x1 + x2 + x3 + 
                         x1*x2 + x1*x3 + x2*x3 +
                         I(x1^2) + I(x2^2) + I(x3^2), data = dX)
    
  }else if (n_factor == 4){
    dX <- data.frame(matrix(Xvec, 
                            byrow = F, 
                            nrow = N, 
                            ncol = n_factor))
    names(dX) <- c("x1", "x2", "x3", "x4")
    X <- model.matrix(~ 1 + x1 + x2 + x3 + x4 + 
                        x1*x2 + x1*x3 + x1*x4 + x2*x3 + x2*x4 + x3*x4 + 
                        I(x1^2) + I(x2^2) + I(x3^2) + I(x4^2), data = dX)
  }else if (n_factor == 5){
    dX <- data.frame(matrix(Xvec, 
                            byrow = F, 
                            nrow = N, 
                            ncol = n_factor))
    names(dX) <- c("x1", "x2", "x3", "x4", "x5")
    X <- model.matrix(~ 1 + x1 + x2 + x3 + x4 + x5 +
                        x1*x2 + x1*x3 + x1* x4 + x1*x5 + x2*x3 + x2*x4 + x2*x5 +
                        x3*x4 + x3*x5 + x4*x5 + 
                        I(x1^2) + I(x2^2) + I(x3^2) + I(x4^2) + I(x5^2), data = dX)
    
  }else {
    stop("We aren't studying that number of factors!")
  }
  return(X)
}

# 2.) Make  grid on the design space in order to search
#      for the max SPV for a candidate design
#    NOTE:
#    a) Eg Xg_onefac is such a grid for K = 1 design factors
#    b.) We here apply the 5^K grid as applied by 
#        Borkowski (2003) and Hernandez and Nachtsheim (2018)

# number of points in grid in 1 dimension
n <- 5
## K = 1  ---
x1        <- as.list(seq(-1,1,length.out = n))
Xs        <- lapply(x1, function(X) rbind(c(1,X,X^2)))
Xs_onefac <- do.call(rbind, Xs)


## K = 2 ---
x1      <- seq(-1,1,length.out = n)
x2      <- seq(-1,1,length.out = n)
Xg      <- expand.grid(int = 1, x1 = x1, x2= x2)

## NOTE: 
#  when building the expanded design matrix from model.matrix,
#   the sqared terms come first, then the interactions.
Xg$x1sq      <- Xg$x1^2
Xg$x2sq      <- Xg$x2^2
Xg$x1x2      <- Xg$x1*Xg$x2
Xs_twofac    <- as.matrix(Xg)



## K = 3 ---
x1      <- seq(-1,1,length.out = n)
x2      <- seq(-1,1,length.out = n)
x3      <- seq(-1,1,length.out = n)
Xg      <- expand.grid(int = 1, x1 = x1, x2= x2, x3 = x3)

Xg$x1sq     <- Xg$x1^2
Xg$x2sq     <- Xg$x2^2
Xg$x3sq     <- Xg$x3^2
Xg$x1x2     <- Xg$x1*Xg$x2
Xg$x1x3     <- Xg$x1*Xg$x3
Xg$x2x3     <- Xg$x2*Xg$x3
Xs_threefac <- as.matrix(Xg)

## K = 4 ---
x1      <- seq(-1,1,length.out = n)
x2      <- seq(-1,1,length.out = n)
x3      <- seq(-1,1,length.out = n)
x4      <- seq(-1,1,length.out = n)
Xg      <- expand.grid(int = 1, x1 = x1, x2= x2, x3 = x3, x4 = x4)

Xg$x1sq     <- Xg$x1^2
Xg$x2sq     <- Xg$x2^2
Xg$x3sq     <- Xg$x3^2
Xg$x4sq     <- Xg$x4^2
Xg$x1x2     <- Xg$x1*Xg$x2
Xg$x1x3     <- Xg$x1*Xg$x3
Xg$x1x4     <- Xg$x1*Xg$x4
Xg$x2x3     <- Xg$x2*Xg$x3
Xg$x2x4     <- Xg$x2*Xg$x4
Xg$x3x4     <- Xg$x3*Xg$x4
Xs_fourfac <- as.matrix(Xg)

## K = 5 ---
x1      <- seq(-1,1,length.out = n)
x2      <- seq(-1,1,length.out = n)
x3      <- seq(-1,1,length.out = n)
x4      <- seq(-1,1,length.out = n)
x5      <- seq(-1,1,length.out = n)
Xg      <- expand.grid(int = 1, x1 = x1, x2= x2, x3 = x3, x4 = x4, x5 = x5)

Xg$x1sq     <- Xg$x1^2
Xg$x2sq     <- Xg$x2^2
Xg$x3sq     <- Xg$x3^2
Xg$x4sq     <- Xg$x4^2
Xg$x5sq     <- Xg$x5^2
Xg$x1x2     <- Xg$x1*Xg$x2
Xg$x1x3     <- Xg$x1*Xg$x3
Xg$x1x4     <- Xg$x1*Xg$x4
Xg$x1x5     <- Xg$x1*Xg$x5
Xg$x2x3     <- Xg$x2*Xg$x3
Xg$x2x4     <- Xg$x2*Xg$x4
Xg$x2x5     <- Xg$x2*Xg$x5
Xg$x3x4     <- Xg$x3*Xg$x4
Xg$x3x5     <- Xg$x3*Xg$x5
Xg$x4x5     <- Xg$x4*Xg$x5
Xs_fivefac <- as.matrix(Xg)

# 3.) This function will evaluate the G-eff, Eq (6) in paper by anonymous 
#     author, for a candidate design matrix

G_eff <- function(Xvec, N, n_factor){
  # design points come in as Vec(X), put into
  #   expanded design matrix
  X    <- exp_Design_mat(Xvec, N = N, n_factor = n_factor)
  p    <- ncol(X)
  xpx  <- t(X)%*%X
  
  if(det(xpx) < .Machine$double.eps){
    result   <- -Inf
  }else{
    
    if(n_factor == 1){
      #U   <- chol(xpx)
      #Z   <- forwardsolve(t(U), t(Xs_onefac))
      #D   <- colSums(N*Z^2)
      H <- Xs_onefac%*%solve(xpx, t(Xs_onefac), tol = 1E-20)
      D <- diag(N*H)
    }else if(n_factor == 2){
      # Cholesky decomp gives speed increase in case of
      #  more dense grid
      U   <- chol(xpx)
      Z   <- forwardsolve(t(U), t(Xs_twofac))
      D   <- colSums(N*Z^2)
    }else if(n_factor == 3){
      U   <- chol(xpx)
      Z   <- forwardsolve(t(U), t(Xs_threefac))
      D   <- colSums(N*Z^2)
    }else if(n_factor == 4){
      U   <- chol(xpx)
      Z   <- forwardsolve(t(U), t(Xs_fourfac))
      D   <- colSums(N*Z^2)
    }else if(n_factor == 5){
      U   <- chol(xpx)
      Z   <- forwardsolve(t(U), t(Xs_fivefac))
      D   <- colSums(N*Z^2)
    }
    
    # get the approximation to the max prediction variance
    max.diag <- max(D)
    # return the optimality score of this matrix on the 
    #  G-efficiency scale
    result   <- 100*p/max.diag
  }
  return(result)
}

# =============================================================================
# II) Input and clean the data, construct R-structure matrices
#      for the optimal designs
#
# 1.)  First, we need to read in the reported designs and organize them
# into an R-matrix. The following code accomplishes this task
D <- read.csv("gOpt_designs_via_PSO_for_Validation.csv", stringsAsFactors = F)
# unknown trouble with the first column name
names(D)
names(D)[1] <- "K"

# turn into list for easy apply calcution
L <- list()
for(i in 1:nrow(D)){
  L[[i]] <- D[i,]
}

# 2.) Now, the proposed designs are coded as a Matlab structured matrix,
#  so the following function will turn them into an R matrix so that
#  the Geff can be calculated.



makeR_matrix <- function(dataRow){
  # pick of the G-PSO design
  t1 <- dataRow$G.PSO_Design
  nc <- nchar(t1)
  t1 <- substr(t1, start = 2, stop = nc-1)
  ot <- strsplit(t1, split = c(";"))[[1]]
  MM <- matrix(NA, nrow = dataRow$N, ncol = dataRow$K)
  for(i in 1:length(ot)){
    ttt <- ot[i]
    spt <- as.numeric(strsplit(ttt, split = " ")[[1]])
    spt2 <- spt[!is.na(spt)]
    MM[i, ] <- spt2
  }
  # store back in design scenario, make a sublist
  #  where the R design matrix is the second element
  Ltemp <- list(pso_data = dataRow, pso_design = MM)
  return(Ltemp)
}

## apply the makeR_matrix function to each list element
L <- lapply(L, makeR_matrix)

##==============================================================================
#  III) REPRODUCE G-EFFICIENCIES OF DESIGNS REPORTED IN PAPER
#      * this confirms the G-optimality of the proposed designs
#        and is sufficient to demonstrate that we have found
#        improved designs.

getGeff_list <- function(data_l){
  design <- data_l$pso_design
  Ktemp  <- data_l$pso_data$K
  Ntemp  <- data_l$pso_data$N
  
  Geff_temp <- G_eff(Xvec = design, 
                     N = Ntemp, 
                     n_factor = Ktemp)
  return(data.frame(K = Ktemp, N = Ntemp, Geff_validation = Geff_temp))
}

test <- L[[29]]$pso_design

Geff_list <- lapply(L, getGeff_list)

Geff_PSO_validation_data <- do.call(rbind, Geff_list)

##==============================================================================
#  IV) VALIDATION, COMPARE COMPUTED GEFFS FOR DESIGNS REPORTED TO THE 
#      GEFF REPORETED IN THE PAPER
#    NOTE: This reproduces the reported data in Table 3 for the K = 4,5 
#           design scenarios, and, for the K = 1,2,3, the Geffs reported
#           here correspond to the best designs illustrated in Figure 1,
#           to reproduce the data in Table 1, we would have to load
#           the GA-designs of Borkowski (2003), compute their G-effs,
#           and then compute the relative G-releffs of the G-PSO designs
#           to the G-GA designs.  This can be computed by hand for a few
#           if you take the reported G-effs of Borkowski (2003)


# gram the newly computed G-eff and the reported G-eff and compare
computed_Geff <- Geff_PSO_validation_data$Geff_validation
reported_Geff <- D$Publishhed_G.eff

print(cbind(computed_Geff, reported_Geff))

# note that there is some error at the order of E-9, I believe
#  this is due to computing these scores as reported in the paper
#  in a different programming language
print(round(computed_Geff - reported_Geff,2))
# validate!
print(all.equal(computed_Geff, reported_Geff))



